/*******************************************************************************

FILE:           Otschi.c

TYPE:           C source file

PLATFORMS:      Win32

DESCRIPTION:    Main source file for the Otschi application

COPYRIGHT:      (c) 2002-2005 Francesco Trotta

VERSION:        3.20

*******************************************************************************/

// Windows header files:
#define WIN32_LEAN_AND_MEAN // Exclude rarely-used stuff from Windows headers.
#include <windows.h>
#include <windowsx.h>

// Many older and newer compilers ship with incomplete Windows headers.

#ifndef CS_DROPSHADOW
#define CS_DROPSHADOW 0x00020000
#endif // #ifndef CS_DROPSHADOW

#ifndef LR_SHARED
#define LR_SHARED 0x8000
#endif // #ifndef LR_SHARED

#ifndef MONITOR_DEFAULTTOPRIMARY
#define MONITOR_DEFAULTTOPRIMARY 0x00000001
#endif // #ifndef MONITOR_DEFAULTTOPRIMARY

// Define required multiple display monitor data types if not yet.
#if \
    defined __LCC__ || \
    (defined __BORLANDC__ && WINVER < 0x0500) || \
    (defined _MSC_VER && WINVER < 0x0500)
DECLARE_HANDLE(HMONITOR);
typedef struct tagMONITORINFO {
    DWORD   cbSize;
    RECT    rcMonitor;
    RECT    rcWork;
    DWORD   dwFlags;
} MONITORINFO, *LPMONITORINFO;
#endif

// Make sure the INT_PTR and UINT_PTR data types are defined.
#if defined __BORLANDC__ && WINVER < 0x0500
typedef int INT_PTR;
typedef UINT UINT_PTR;
#endif

// Make sure the LONG_PTR data type is defined.
#if \
    (defined __BORLANDC__ && WINVER < 0x0500) || \
    (defined _MSC_VER && WINVER < 0x0500)
typedef long LONG_PTR;
#endif

/* Replace GetWindowLongPtr and SetWindowLongPtr with more type-safe inline
 * functions on Win32 compilers. */

#ifndef _WIN64

#undef GetWindowLongPtr
#define GetWindowLongPtr(hWnd, nIndex) (xGetWindowLongPtr(hWnd, nIndex))
LONG_PTR xGetWindowLongPtr(HWND hWnd, int nIndex)
{ return (LONG_PTR)GetWindowLong(hWnd, nIndex); }

#undef SetWindowLongPtr
#define SetWindowLongPtr(hWnd, nIndex, dwNewLong) \
    (xSetWindowLongPtr(hWnd, nIndex, dwNewLong))
LONG_PTR xSetWindowLongPtr(HWND hWnd, int nIndex, LONG_PTR dwNewLong)
{ return (LONG_PTR)SetWindowLong(hWnd, nIndex, (LONG)dwNewLong); }

#endif // #ifndef _WIN64

// Special macros:

#ifdef _MSC_VER
#define DEFAULT_ASSUME_0 default: __assume(0);
#else
#define DEFAULT_ASSUME_0
#endif

// Application specific header files:

#include "Resource.h"
#include "RIG.h"



/*******************************************************************************
*                                                                              *
*                        Global Data Types & Constants                         *
*                                                                              *
*******************************************************************************/

// Number of colors used to paint an Otschi face.
#define OTSCHICOLORCOUNT    4

// Metric constants:
#define OTM_DIAM            96      // diameter of the Otschi face circle
#define OTM_BLINKX          30      // left of the blinking rectangle
#define OTM_BLINKY          24      // top of the blinking rectangle
#define OTM_BLINKW          36      // width of the blinking rectangle
#define OTM_BLINKH          19      // height of the blinking rectangle

// Blinking time constants, in milliseconds:
#define OTBT_LONGAUFMIN     2500    // minimum LONGAUF interval
#define OTBT_LONGAUFMAX     3500    // maximum LONGAUF interval
#define OTBT_LONGZUMIN      1500    // minimum LONGZU interval
#define OTBT_LONGZUMAX      2000    // maximum LONGZU interval
#define OTBT_SHORT          50      // SHORTAUF and SHORTZU interval

typedef UINT OTSCHIBLIPER;

/*------------------------------------------------------------------------------
The OTSCHIBLIPER (bliper = blinking period) flags are used to identify the times
while the eyes remain either open or closed. There are four blipers repeating in
the following sequence:

Bliper name     Value   Description

LONGAUF         0x00    eyes open for a longer time
LONGZU          0x01    eyes closed for a longer time
SHORTAUF        0x02    eyes open for a shorter time
SHORTZU         0x03    eyes closed for a shorter time

Starting with version 3.20, the bliper of an otschi face is stored in the lower
3 bits of the owned window user data (bit 2 is reserved and must be 0). This
value always refers to the NEXT bliper in sequence, not to the current one.
------------------------------------------------------------------------------*/

#define OTBP_ZU             0x01            // the eyes are closed
#define OTBP_SHORT          0x02            // the period is short
#define OTBP_LONGAUF        0x00
#define OTBP_LONGZU         OTBP_ZU
#define OTBP_SHORTAUF       OTBP_SHORT
#define OTBP_SHORTZU        (OTBP_ZU | OTBP_SHORT)
#define OTBP_LAST           OTBP_SHORTZU    // last status before recycling

/*------------------------------------------------------------------------------
The following enumeration is used to control the action taken by PaintProc after
selecting and realizing the Otschi palette in the DC, and to determine the way
the graphics parameter should be used.
------------------------------------------------------------------------------*/

typedef enum
{ PPA_BITBLT, PPA_REALIZEPALETTE }
PAINTACTION;

/* This window message may be sent to any Otschi window to open a new Otschi
 * face in the context of the existing thread: */
#define WM_OPENOTSCHI (WM_APP + 20)



/*******************************************************************************
*                                                                              *
*                                 Global Data                                  *
*                                                                              *
*******************************************************************************/

// Global constants and pseudo-constants:
static ATOM        _atom;
static const TCHAR _szWindowClass[] = TEXT("OTSCHI 3.20");  // window class name
static const TCHAR _szAppTitle[]    = TEXT("Otschi");       // program name
static HDC         _hdcMem;                                 // memory DC
static HINSTANCE   _hInstance;                              // current instance
static HPALETTE    _hpal;                                   // palette

// cursor position on window
static int _nXCur, _nYCur;

// global face counter
static unsigned int _cOtschis = 0;



/*******************************************************************************
*                                                                              *
*                            Function Declarations                             *
*                                                                              *
*******************************************************************************/

ATOM                RegisterClassApp(HINSTANCE);
BOOL                IsOtschiWindow(HWND);
BOOL                WorkAreaFromXY(int, int, PRECT);
HDC                 InitDC(HINSTANCE);
HINSTANCE           InitInstance(HINSTANCE);
HWND                CreateUnownedWindow(LPCTSTR, DWORD);
int                 AdjustOwnedCoord(int, const LONG [3]);
int                 SubMain(HINSTANCE);
LRESULT             OnCreate(HWND, LPCREATESTRUCT);
LRESULT CALLBACK    WindowProc(HWND, UINT, WPARAM, LPARAM);
OTSCHIBLIPER        GetBliper(HWND hwndOwned);
void                BlinkOtschi(HWND, UINT_PTR);
void                DestroyOtschi(HWND);
void                DragOtschi(HWND, const POINTS);
void                PaintProc(HWND, PAINTACTION);
void                ShowErrorMessage(UINT);

/*------------------------------------------------------------------------------
The OtschiMain function replaces WinMain when OTSCHI_MAIN is defined, producing
smaller executable code. Note that you also need to set the application entry
point properly in the linker settings when using this feature.
------------------------------------------------------------------------------*/
#ifdef OTSCHI_MAIN
void __stdcall      OtschiMain(void);
#endif



/*******************************************************************************
*                                                                              *
*                             Function Definitions                             *
*                                                                              *
*******************************************************************************/

/*******************************************************************************

FUNCTION:   AdjustOwnedCoord

PURPOSE:    Corrects a screen coordinate (x or y) for the location of an owned
            Otschi window using the bounds of the display monitor's work area.

PARAMETERS:
 coord
    The screen coordinate being corrected.
 bounds
    Contains the left- or topmost coordinate of the monitor's work area in
    bounds[0]; the right- or bottommost coordinate is in bounds[2].

RETURNS:    The corrected coordinate value.

*******************************************************************************/

int AdjustOwnedCoord(int coord, const LONG bounds[3])
{
    {
        int bound = bounds[0];
        if (coord < bound) return bound;
    }

    {
        int bound = bounds[2] - OTM_DIAM;
        if (coord > bound) return bound;
    }

    return coord;
}

/*******************************************************************************

FUNCTION:   BlinkOtschi

PURPOSE:    Performs all necessary operations to let the eyes blink on the owned
            Otschi window.
            Blinking is the process of switching a face from a period in which
            the eyes are open to a period where they are closed and vice-versa.

PARAMETERS:
 hwndOwned
    Handle to an owned Otschi window.
 nIDEvent
    Identifier of the timer currently associated with the Otschi window, or 0 if
    no such timer exists.

*******************************************************************************/

void BlinkOtschi(HWND hwndOwned, UINT_PTR nIDEvent)
{
    {
        UINT uElapse;

        OTSCHIBLIPER otbp = GetBliper(hwndOwned);

        // It is not safe to assume (dwOtbp & OTBP_LAST) == dwOtbp!
        switch (otbp & OTBP_LAST)
        {
        case OTBP_LONGAUF:
            uElapse =
                OTBT_LONGAUFMIN + GRIG_Next(OTBT_LONGAUFMAX - OTBT_LONGAUFMIN);
            break;
        case OTBP_LONGZU:
            uElapse =
                OTBT_LONGZUMIN + GRIG_Next(OTBT_LONGZUMAX - OTBT_LONGZUMIN);
            break;
        case OTBP_SHORTAUF:
        case OTBP_SHORTZU:
            uElapse = OTBT_SHORT;
            break;
        DEFAULT_ASSUME_0
        }

        SetTimer(hwndOwned, nIDEvent, uElapse, NULL);

        SetWindowLongPtr(
            hwndOwned,
            GWL_USERDATA,
            (otbp + 1) & ~(OTBP_LAST + 1)
            );
    }

    {
        // RECT with boudaries of the eyes
        static const RECT _rcBlinkUpdate =
        {
            OTM_BLINKX,
            OTM_BLINKY,
            OTM_BLINKX + OTM_BLINKW,
            OTM_BLINKY + OTM_BLINKH
        };

        InvalidateRect(hwndOwned, &_rcBlinkUpdate, FALSE);
    }
}

/*******************************************************************************

FUNCTION:   CreateUnownedWindow

PURPOSE:    Creates an Otschi window without an owner, that is the main window
            or an owner window.

PARAMETERS:
 lpszWindow
    Pointer to a null-terminated string that specifies the window name.
 dwStyle
    The style of the window being created.

RETURNS:    A handle to the new window on success, otherwise NULL.

*******************************************************************************/

/*------------------------------------------------------------------------------
There are some important requirements on the Otschi application that determine
the whole window hierarchy design:

- Only one entry in the task manager
- No taskbar button
- Automatic position adjustment when display size or location changes

A taskbar button is created by the system for every visible, top-level window
that doesn't have the extended style WS_EX_TOOLWINDOW: windows without this
style are moved automatically by the system when a display setting such as the
screen resolution or location on the virtual screen of a display monitor
changes. Now we would like to allow automatic position adjustment for Otschi
faces but we don't want any taskbar buttons. To accomplish that, we use a double
window system consisting of an invisible owner window with the WS_EX_TOOLWINDOW
extended style set and a visible owned window without that style. Owned windows
never appear in the taskbar, not even when the WS_EX_TOOLWINDOW extended style
is not set.
In our implementation, the entry in the task manager is accomplished through a
zero-sized yet visible top-level window referred to as the main window.
------------------------------------------------------------------------------*/

HWND CreateUnownedWindow(LPCTSTR lpszWindow, DWORD dwStyle)
{
    // Create owner:
    return
        CreateWindowEx(
        WS_EX_TOOLWINDOW, // WS_EX_TOOLWINDOW for no taskbar button!
        _szWindowClass,
        lpszWindow,
        dwStyle,
        0,
        0,
        0,
        0,
        NULL,
        NULL,
        _hInstance,
        NULL
        );
}

/*******************************************************************************

FUNCTION:   DestroyOtschi

PURPOSE:    Destroys an owned Otschi window and its owner.

PARAMETERS:
 hwndOwned
    Handle to an owned Otschi window.

NOTES:      Always call this function to remove a single Otschi face from the
            screen. This function automatically updates the global face counter
            and posts WM_QUIT when the last face is closed.

*******************************************************************************/

void DestroyOtschi(HWND hwndOwned)
{
    DestroyWindow(GetWindow(hwndOwned, GW_OWNER));
    if (--_cOtschis == 0) PostQuitMessage(0);
}

/*******************************************************************************

FUNCTION:   DragOtschi

PURPOSE:    Moves an owned Otschi window to a new position on the screen as the
            user drags it with the mouse.

PARAMETERS:
 hwndOwned
    Handle to an owned Otschi window.
 ptCursor
    The coordinates of the cursor on the window as a POINTS structure.

*******************************************************************************/

void DragOtschi(HWND hwndOwned, const POINTS ptCursor)
{
    int nX, nY;
    RECT rc;

    if (!GetWindowRect(hwndOwned, &rc)) return;

    nX = rc.left    + ptCursor.x;
    nY = rc.top     + ptCursor.y;

    if (!WorkAreaFromXY(nX, nY, &rc)) return;

    SetWindowPos(
        hwndOwned,
        NULL,
        AdjustOwnedCoord(nX - _nXCur, &rc.left),
        AdjustOwnedCoord(nY - _nYCur, &rc.top),
        0,
        0,
        SWP_NOSIZE          |
        SWP_NOZORDER        |   // Save a message.
        SWP_NOSENDCHANGING      // Save another message.
        );
}

/*******************************************************************************

FUNCTION:   GetBliper

PURPOSE:    Retrieves the OTSCHIBLIPER value of a face.

PARAMETERS:
 hwndOwned
    Handle to the face's owned window.

RETURNS:    The OTSCHIBLIPER of the face with the specified owned window.

NOTES:      Although this function may fail, no special value is reserved to
            indicate an error. Instead, if an error occurs, the return value is
            undefined.

*******************************************************************************/

OTSCHIBLIPER GetBliper(HWND hwndOwned)
{ return (OTSCHIBLIPER)GetWindowLongPtr(hwndOwned, GWL_USERDATA); }

/*******************************************************************************

FUNCTION:   InitDC

PURPOSE:    Creates and initializes the memory DC and the palette. The Otschi
            face with open eyes is drawn in the upper half of the DC; the face
            with closed eyes is drawn in the lower half of the DC.

PARAMETERS:
 hInstance
    Handle to the current module instance.

RETURNS:    If the function succeeds, the return value is the handle to the
            memory DC stored in _hdcMem. If the function fails, the return value
            is NULL.

NOTES:      Since this function is called just one time upon program
            initialization, no cleanup operation is performed in case of
            failure.

*******************************************************************************/

HDC InitDC(HINSTANCE hInstance)
{
    HDC hdcDest, hdcSrc;
    HGDIOBJ hgdiobj, hOldSrcBitmap;

    // 1st step: Create source DC and select IDB_OTSCHI bitmap into it.

    hdcSrc = CreateCompatibleDC(NULL);

    hOldSrcBitmap =
        SelectObject(
        hdcSrc,
        (HBITMAP)
        LoadImage(
        hInstance,
        MAKEINTRESOURCE(IDB_OTSCHI),
        IMAGE_BITMAP,
        0,
        0,
        LR_CREATEDIBSECTION)
        );

    if (hOldSrcBitmap == NULL) return NULL;


    // 2nd step: Create palette object from logical palette structure.

    {
        /* In previous versions, the color entries were retrieved directly from
         * the source DC using the GetDIBColors function. As it turns out, that
         * trick didn't work well on historical Windows 3.51. In order to let
         * Otschi run on Windows 3.51, version 3.20 keeps a copy of the color
         * table in-code. */

        static const struct
        {
            WORD         palVersion;
            WORD         palNumEntries;
            PALETTEENTRY palPalEntry[OTSCHICOLORCOUNT];
        }
        _logpal =
        {
            0x0300,
            OTSCHICOLORCOUNT,
            {
                { 0xff, 0xc0, 0xc0, 0x00 },
                { 0x00, 0x00, 0x00, 0x00 },
                { 0xff, 0x00, 0x00, 0x00 },
                { 0x84, 0x00, 0x00, 0x00 },
            }
        };

        HPALETTE hpal;

        hpal = CreatePalette((CONST LOGPALETTE *)&_logpal);
        if (hpal == NULL) return NULL;
        _hpal = hpal;
    }

    /*
     * 3rd step: Create destination DC and select a compatible bitmap as wide
     * and twice as high as IDB_OTSCHI bitmap into it. Then copy the source DC
     * twice into the destination DC (up and down).
     */

    hdcDest = CreateCompatibleDC(NULL);

    hgdiobj =
        SelectObject(
        hdcDest,
        CreateCompatibleBitmap(hdcSrc, OTM_DIAM, 2 * OTM_DIAM));

    if (hgdiobj == NULL) return NULL;

    if (!BitBlt(hdcDest, 0, 0, OTM_DIAM, OTM_DIAM, hdcSrc, 0, 0, SRCCOPY))
        return NULL;

    if (
        !BitBlt(
        hdcDest,
        0,
        OTM_DIAM,
        OTM_DIAM,
        OTM_DIAM,
        hdcSrc,
        0,
        0,
        SRCCOPY))
        return NULL;

    /*
     * 4th step: Select IDB_OTSCHIZU into the source DC and copy it to the lower
     * half of the destination DC.
     */

    hgdiobj =
        SelectObject(
        hdcSrc,
        (HBITMAP)
        LoadImage(
        hInstance,
        MAKEINTRESOURCE(IDB_OTSCHIZU),
        IMAGE_BITMAP,
        0,
        0,
        LR_CREATEDIBSECTION)
        );

    if (hgdiobj == NULL) return NULL;

    // Delete IDB_OTSCHI bitmap (previously selected in the source DC).
    DeleteObject(hgdiobj);

    if (
        !BitBlt(
        hdcDest,
        OTM_BLINKX,
        OTM_DIAM + OTM_BLINKY,
        OTM_BLINKW,
        OTM_BLINKH,
        hdcSrc,
        0,
        0,
        SRCCOPY))
        return NULL;

    /*
     * Final step: Delete source DC and its bitmap and store destination DC in
     * global pseudo-constant.
     */

    DeleteObject(SelectObject(hdcSrc, hOldSrcBitmap));
    DeleteDC(hdcSrc);

    return _hdcMem = hdcDest;
}

/*******************************************************************************

FUNCTION:   InitInstance

PURPOSE:    Performs any necessary initialization and shows the first Otschi
            face.

PARAMETERS:
 hInstance
    Handle to the current module instance.

RETURNS:    If the function succeeds, the return value is hInstance.
            If the function fails, the return value is NULL.

NOTES:      Since this function is called just one time upon program
            initialization, no cleanup operation is performed in case of
            failure.

*******************************************************************************/

HINSTANCE InitInstance(HINSTANCE hInstance)
{
    // Initialize memory DC:
    if (InitDC(hInstance) == NULL) return NULL;

    // Register application window class:
    {
        ATOM atom = RegisterClassApp(hInstance);
        if (atom == 0) return NULL;
        _atom = atom;
    }

    // Randomization:
    GRIG_Randomize();

    // Set thread priority below normal:
    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);

    // Create main window:

    // WS_VISIBLE style for entry in task manager!
    // WS_POPUP for no border!
    if (CreateUnownedWindow(_szAppTitle, WS_VISIBLE | WS_POPUP) == NULL)
        return NULL;

    // Create the first owner window which in turn creates the owned window:
    if (CreateUnownedWindow(NULL, WS_POPUP) == NULL) return NULL;

    return hInstance;
}

/*******************************************************************************

FUNCTION:   IsOtschiWindow

PURPOSE:    Determines if a specified window identifies an Otschi window
            belonging to the current process.

PARAMETERS:
 hwnd
    Handle to the window to test.

RETURNS:    nonzero if the specified window is an otschi window belonging to the
            current thread; otherwise, zero.

*******************************************************************************/

BOOL IsOtschiWindow(HWND hwnd)
{ return (ATOM)GetClassLong(hwnd, GCW_ATOM) == _atom; }

/*******************************************************************************

FUNCTION:   main

PURPOSE:    The main function is never called, it is only needed to properly
            link the app under some circumstances when the entry point is set to
            OtschiMain. Visual C++ raises an error under the debug configuration
            when this function is missing. The gcc compiler also requires a main
            function when compiling without the option "-nostartfiles". Both
            OTSCHI_MAIN and _MAIN must be defined for the main function to be
            included in the link.

*******************************************************************************/

#if defined(OTSCHI_MAIN) && defined(_MAIN)
void __cdecl main() {}
#endif

/*******************************************************************************

FUNCTION:   OnCreate

PURPOSE:    Processes the WM_CREATE message for all kinds of Otschi windows.

PARAMETERS:
 hWnd
    Handle to the window that received the message.
 lpCreateStruct
    (LPCREATESTRUCT)lParam

RETURNS:    A return value for the WindowProc.

*******************************************************************************/

LRESULT OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
    // IF owner Otschi window...
    if (!(lpCreateStruct->style & WS_VISIBLE))
    {
        RECT rc;

        // Calculate the owned window position:

        /* Retrieve the work area rectangle of the primary display monitor, then
         * create owned window. */
        if (
            WorkAreaFromXY(0, 0, &rc)

            &&

            CreateWindowEx(
            0,
            _szWindowClass,
            NULL,
            WS_VISIBLE | WS_POPUP, // WS_POPUP for no border!
            (rc.right - rc.left - OTM_DIAM) / 2,
            (rc.bottom - rc.top - OTM_DIAM) / 2,
            OTM_DIAM,
            OTM_DIAM,
            hwnd,
            NULL,
            lpCreateStruct->hInstance,
            NULL)
            !=
            NULL)

            return 0;
    }

    // IF main Otschi window...
    else if (lpCreateStruct->hwndParent == NULL)
        return 0;

    // IF owned Otschi window...
    else
    {
        /* Create window region: note that the lower-right corner has the
         * coordinates (OTM_DIAM + 1, OTM_DIAM + 1). */
        HRGN hrgn = CreateEllipticRgn(0, 0, OTM_DIAM + 1, OTM_DIAM + 1);
        if (hrgn != NULL)
        {
            if (SetWindowRgn(hwnd, hrgn, FALSE))
            {
                ++_cOtschis; // global counter only increased on success
                BlinkOtschi(hwnd, 0);
                return 0;
            }
            else
                DeleteObject(hrgn);
        }
    }

    return -1; // destroy mistaken window
}

/*******************************************************************************

FUNCTION:   OtschiMain

PURPOSE:    This function is intended to replace WinMain whenever the linker
            settings allow a user-defined enrty point, in order to generate
            smaller executable code. Define the symbol OTSCHI_MAIN in your
            compiler's settings to hide the WinMain definition and enable
            OtschiMain. You may also need to define the symbol _MAIN to enable
            the function main.

*******************************************************************************/

#ifdef OTSCHI_MAIN
void __stdcall OtschiMain(void)
{
    HINSTANCE hInstance;

    hInstance = GetModuleHandle(NULL);

    /* Nothing can be done without an instance handle, not even load an error
     * message! Must call ExitProcess, otherwise the process may keep staying
     * resident forever; the standard entry calls ExitProcess automatically. */
    ExitProcess(hInstance == NULL ? 0 : SubMain(hInstance));
}
#endif // #ifdef OTSCHI_MAIN

/*******************************************************************************

FUNCTION:   PaintProc

PURPOSE:    Performs graphical operations on the owned Otschi window.

PARAMETERS:
 hwndOwned
    Handle to an owned Otschi window.
 action
    Must specify one of the following values:
     PPA_BITBLT
        Indicates that all or part of the client area must be repainted. This
        value must be specified when calling the function in response to a
        WM_PAINT message.
     PPA_REALIZEPALETTE
        Maps the Otschi logical palette to the system palette and the
        invalidates the client area.

NOTES:      This function is called in response to a window message to ensure
            the system palette gets updated before further action is performed.

*******************************************************************************/

void PaintProc(HWND hwndOwned, PAINTACTION action)
{
    HDC hdc;
    PAINTSTRUCT paint;

    switch (action)
    {
        case PPA_BITBLT:
            hdc = BeginPaint(hwndOwned, &paint);
            break;
        case PPA_REALIZEPALETTE:
            hdc = GetDC(hwndOwned);
            break;
        DEFAULT_ASSUME_0
    }

    if (hdc != NULL)
    {
        HPALETTE hpalOld;

        // Select Otschi palette:
        hpalOld = SelectPalette(hdc, _hpal, FALSE);

        switch (action)
        {
        case PPA_BITBLT:
            {
                int x, y, width, height;
                RECT *prcPaint = &(paint.rcPaint);

                x       = prcPaint->left;
                y       = prcPaint->top;
                width   = prcPaint->right - x;
                height  = prcPaint->bottom - y;

                if (width > 0 && height > 0)
                {
                    BitBlt(
                        hdc,
                        x,
                        y,
                        width,
                        height,
                        _hdcMem,
                        x,
                        y + (GetBliper(hwndOwned) & OTBP_ZU ? 0 : OTM_DIAM),
                        SRCCOPY
                    );
                }
            }
            break;
        case PPA_REALIZEPALETTE:
            /*
             * UnrealizeObject is only called for compliance with Windows
             * programming guidelines. I just can't see why it's there. If this
             * call fails, we simply go on and forget about that.
             * UnrealizeObject must be called before RealizePalette!
             */
            UnrealizeObject(_hpal);
            if (hpalOld != NULL)
            {
                UINT uRes = RealizePalette(hdc);
                if (uRes != 0 && uRes != GDI_ERROR)
                    InvalidateRect(hwndOwned, NULL, FALSE);
            }
            break;
        DEFAULT_ASSUME_0
        }

        // Restore old palette:
        if (hpalOld != NULL) SelectPalette(hdc, hpalOld, FALSE);

        switch (action)
        {
        case PPA_BITBLT:
            EndPaint(hwndOwned, &paint);
            break;
        case PPA_REALIZEPALETTE:
            ReleaseDC(hwndOwned, hdc);
            break;
        DEFAULT_ASSUME_0
        }
    }
}

/*******************************************************************************

FUNCTION:   RegisterClassApp

PURPOSE:    Registers the application window class.

PARAMETERS:
 hInstance
    Handle to the current module instance.

RETURNS:    If the function succeeds, the return value is a class atom of the
            registered window class. If the function fails, the return value is
            0.

NOTES:      Since this function is called just one time upon program
            initialization, no cleanup operation is performed in case of
            failure.

*******************************************************************************/

ATOM RegisterClassApp(HINSTANCE hInstance)
{
    ATOM atom;
    HCURSOR hCursor;
    HICON hIcon, hIconSm;
    WNDCLASSEX wcex;

    // Load icon:
    hIcon =
        (HICON)
        LoadImage(
        hInstance,
        MAKEINTRESOURCE(IDI_OTSCHI),
        IMAGE_ICON,
        0,
        0,
        LR_DEFAULTSIZE
        );
	if (hIcon == NULL) return 0;

    // Load small icon:
    hIconSm =
        (HICON)
        LoadImage(
        hInstance,
        MAKEINTRESOURCE(IDI_OTSCHI),
        IMAGE_ICON,
        GetSystemMetrics(SM_CXSMICON),
        GetSystemMetrics(SM_CYSMICON),
        0
        );
    if (hIconSm == NULL) return 0;

    // Load cursor:
	hCursor =
        (HCURSOR)LoadImage(NULL, IDC_ARROW, IMAGE_CURSOR, 0, 0, LR_SHARED);
	if (!hCursor) return 0;

    // Set up WNDCLASSEX structure:
    wcex.cbSize         = sizeof wcex;
    wcex.style          = CS_CLASSDC | CS_DROPSHADOW;
    wcex.lpfnWndProc    = WindowProc;
    wcex.cbClsExtra     = 0;
    wcex.cbWndExtra     = 0;
    wcex.hInstance      = hInstance;
    wcex.hIcon          = hIcon;
    wcex.hCursor        = hCursor;
    wcex.hbrBackground  = NULL;
    wcex.lpszClassName  = _szWindowClass;
    wcex.lpszMenuName   = NULL;
    wcex.hIconSm        = hIconSm;

    // Register window class:
    atom = RegisterClassEx(&wcex);
    if (atom != 0) return atom;

    /* Retry registration without CS_DROPSHADOW style (introduced with Windows
     * XP). */
    wcex.style = 0;
    return RegisterClassEx(&wcex);
}

/*******************************************************************************

FUNCTION:   ShowErrorMessage

PURPOSE:    Shows a message box with an error message to the user.

PARAMETERS:
 uTextId
    Resource identifier of the string with the error message.

NOTES:      This function doesn't return until the message box is closed.

*******************************************************************************/

void ShowErrorMessage(UINT uTextId)
{
    MSGBOXPARAMS mbp;

    mbp.cbSize              = sizeof mbp;
    mbp.hwndOwner           = NULL;
    mbp.hInstance           = _hInstance;
    mbp.lpszText            = (LPCTSTR)(UINT_PTR)uTextId;
    mbp.lpszCaption         = _szAppTitle;
    mbp.dwStyle             = MB_OK | MB_ICONSTOP;
    // mbp.lpszIcon
    // mbp.dwContextHelpId
    mbp.lpfnMsgBoxCallback  = NULL;
    mbp.dwLanguageId        = MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL);

    MessageBoxIndirect(&mbp);
}

/*******************************************************************************

FUNCTION:   SubMain

PURPOSE:    Does everything the program does, except for retrieving the module
            instance. SubMain is called directly by the entry point function
            (WinMain or OtschiMain, depending on the compiler settings).

PARAMETERS:
 hInstance
    Handle to the current module instance.

RETURNS:    The exit value of the program.

*******************************************************************************/

int SubMain(HINSTANCE hInstance)
{
    /* If another instance of Otschi is running, we may just pick any existing
     * Otschi window and send it a WM_OPENOTSCHI message. If this succeeds, we
     * are finished: */
    {
        HWND hwndServer = FindWindow(_szWindowClass, NULL);
        if (hwndServer != NULL)
        {
            HWND hwndOwner =
                (HWND)SendMessage(hwndServer, WM_OPENOTSCHI, 0, 0);
            if (hwndOwner != NULL)
            {
                SetForegroundWindow(hwndOwner);
                return 0;
            }
        }
    }

    _hInstance = hInstance; // required by ShowErrorMessage

    // If no mouse present, show an error message and exit:
    if (GetSystemMetrics(SM_CMOUSEBUTTONS) == 0)
    {
        ShowErrorMessage(IDS_ERR_NOMOUSE);
        return 0;
    }

    // Required initialization:
    if (InitInstance(hInstance) == NULL)
    {
        ShowErrorMessage(IDS_ERR_WINAPI);
        return 0;
    }

    // Message loop:
    for (;;)
    {
        MSG msg;

        switch(GetMessage(&msg, NULL, 0, 0))
        {
        case 0:     // WM_QUIT
            return (int)msg.wParam;
        case -1:    // error
            ShowErrorMessage(IDS_ERR_WINAPI);
            return 0;
        default:    // ok
            /* The task manager may post WM_CLOSE to the main Otschi window to
             * terminate the app. In no other normal circumstance this message
             * is received. */
            if (msg.message == WM_CLOSE) return 0;

            // No TranslateMessage required!
            DispatchMessage(&msg);
        }
    }
}

/*******************************************************************************

FUNCTION:   WinMain

PURPOSE:    The standard Windows application entry point.

NOTES:      This function is only processed by the compiler when the symbol
            OTSCHI_MAIN is not defined. Setting OtschiMain as entry point
            generates less code.

*******************************************************************************/

#ifndef OTSCHI_MAIN
int
APIENTRY WinMain(
    HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    LPTSTR lpCmdLine,
    int nCmdShow)
{

/* The following lines are required to suppress a warning message which is
 * generated by Microsoft and Borland compilers when a function parameter is not
 * used. Borland compilers also warn when a statement has no effect; this may be
 * changed through a #pragma directive. */

#if defined __BORLANDC__ || defined _MSC_VER
#ifdef __BORLANDC__
#pragma warn -eff
#endif // __BORLANDC__
    hPrevInstance; lpCmdLine; nCmdShow;
#ifdef __BORLANDC__
#pragma warn .eff
#endif // __BORLANDC__
#endif // #if defined __BORLANDC__ || defined _MSC_VER

    return SubMain(hInstance);
}
#endif // #ifndef OTSCHI_MAIN

/*******************************************************************************

FUNCTION:   WindowProc

PURPOSE:    Processes messages for the Otschi windows.

*******************************************************************************/

LRESULT CALLBACK WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
    case WM_CREATE:
        return OnCreate(hWnd, (LPCREATESTRUCT)lParam);
    case WM_KEYDOWN:
        if (wParam != VK_ESCAPE) goto def_wnd_proc;
        DestroyOtschi(hWnd);
        break;
    case WM_LBUTTONDOWN:
        _nXCur = GET_X_LPARAM(lParam);
        _nYCur = GET_Y_LPARAM(lParam);
        SetCapture(hWnd);
        break;
    case WM_LBUTTONUP:
        ReleaseCapture();
        break;
    case WM_MOUSEMOVE:
        if (wParam & MK_LBUTTON) DragOtschi(hWnd, MAKEPOINTS(lParam));
        break;
    case WM_OPENOTSCHI:
        return (LRESULT)CreateUnownedWindow(NULL, WS_POPUP);
    case WM_PAINT:
        PaintProc(hWnd, PPA_BITBLT);
        break;
    case WM_PALETTECHANGED:
    case WM_PALETTEISCHANGING:
        if (!IsOtschiWindow((HWND)wParam)) PaintProc(hWnd, PPA_REALIZEPALETTE);
        break;
    case WM_QUERYNEWPALETTE:
        PaintProc(hWnd, PPA_REALIZEPALETTE);
        return TRUE;
    case WM_RBUTTONUP:
        /* Destroying the window on WM_RBUTTONDOWN would let the WM_RBUTTONUP
         * message reach the window below. */
        DestroyOtschi(hWnd);
        break;
    case WM_TIMER:
        BlinkOtschi(hWnd, (UINT_PTR)wParam);
        break;
    default:
def_wnd_proc:
        return DefWindowProc(hWnd, uMsg, wParam, lParam);
    }

    return 0;
}

/*******************************************************************************

FUNCTION:   WorkAreaFromXY

PURPOSE:    Retrieves the work area rectangle of the display monitor that
            contains a specified point.

PARAMETERS:
 x
    Specifies the X location of a point contained in the monitor whose work area
    is being sought, expressed in virtual-screen coordinates.
 y
    Specifies the Y location of a point contained in the monitor whose work area
    is being sought, expressed in virtual-screen coordinates.
 prc
    [out] Pointer to a RECT structure that receives the work area rectangle of
    the specified display monitor, expressed in virtual-screen coordinates.

RETURNS:    Nonzero on success; zero on failure.

NOTES:      If the specified point is not contained within any display monitor,
            this function retrieves the work area rectangle of the primary
            display monitor.

*******************************************************************************/

BOOL WorkAreaFromXY(int x, int y, PRECT prc)
{
    typedef HMONITOR    (WINAPI *MONITORFROMPOINTPROC)(POINT, DWORD);
    typedef BOOL        (WINAPI *GETMONITORINFOPROC)(HMONITOR, LPMONITORINFO);

    static MONITORFROMPOINTPROC _pfnMonitorFromPoint = NULL;
    static GETMONITORINFOPROC   _pfnGetMonitorInfo;

    /* _pfnMonitorFromPoint has two special values: 0 (NULL) to indicate that no
     * function initialization took place yet; 1 to indicate that no multiple
     * display monitor API is available. */
    switch ((INT_PTR)_pfnMonitorFromPoint)
    {
    case 0:
        {
            FARPROC pfn1, pfn2;

            /* If GetModuleHandle fails, GetProcAddress will both fail, too, and
             * the single display monitor routine will be used. */
            HMODULE hModUser32 = GetModuleHandle(TEXT("USER32"));

            if (
                (pfn1 = GetProcAddress(hModUser32, "MonitorFromPoint")) != NULL
                &&
                (pfn2 = GetProcAddress(hModUser32, "GetMonitorInfoA")) != NULL)
            {
                _pfnMonitorFromPoint    = (MONITORFROMPOINTPROC)pfn1;
                _pfnGetMonitorInfo      = (GETMONITORINFOPROC)pfn2;
                break;
            }
        }

        _pfnMonitorFromPoint = (MONITORFROMPOINTPROC)1;
        // don't break...
    case 1:
        return SystemParametersInfo(SPI_GETWORKAREA, 0, prc, 0);
    }

    {
        BOOL bResult;

        {
            MONITORINFO mi;

            mi.cbSize = sizeof mi;
            {
                POINT pt;

                pt.x = x;
                pt.y = y;
                // If MonitorFromPoint fails, GetMonitorInfo will fail, too.
                bResult =
                    _pfnGetMonitorInfo(
                    _pfnMonitorFromPoint(pt, MONITOR_DEFAULTTOPRIMARY),
                    &mi
                    );
            }
            *prc = mi.rcWork; // Do assignement even on failure (why not?).
        }

        return bResult;
    }
}
