/*******************************************************************************

FILE:           RIG.c

TYPE:           C source file

PLATFORMS:      Win32

DESCRIPTION:    RIG is a project independent interface for random 32-bit integer
                generation. This file defines the RIG functions.

COPYRIGHT:      (c) 2002-2005 Francesco Trotta

VERSION:        1.12

*******************************************************************************/

// Windows header files:
#define WIN32_LEAN_AND_MEAN // Exclude rarely-used stuff from Windows headers.
#include <windows.h>

// Source specific header file:
#include "RIG.h"

/*------------------------------------------------------------------------------
The _USEASM macro controls whether optimized inline assembler code should be
used instead of regular C code when possible. This functionality is only tested
under Visual C++, although other compilers may be also able to translate inline
assembler.
------------------------------------------------------------------------------*/

#if defined _M_IX86 && defined _MSC_VER
#define _USEASM
#endif

/*------------------------------------------------------------------------------
The "magic" factors. These numbers generate sequences with period length
MFACT*2^15-1: this is a necessary condition for the algorithm to work properly.
We are using 65184 and 64860. Any two different numbers listed below are just as
fine. The period length of each one of those is a prime number larger than 10^9.
So you can expect a register to run up to 10^18 iterations before the random
sequence repeats.

 65184 64860 64545 64455 64314 64185 63885 63765 63663 63444 63273 63129 62904
 62874 62865 62790 62778 62634 62505 61914 61578 61140 60990 60735 60489 60060
 60009 59775 59325 59253 58758 58443 58368 58149 57918 57780 57774 57603 57525
 57225 57204 57039 56859 56790 56604 56538 56205 55869 55848 55809 55680 55254
 55179 55074 55044 54885 54564 54474 54399 54168 54144 53958 53934 53898 53709
 53529 53199 53130 52950 52743 52509 52275 52215 52023 51933 51873 51813 51744
 51729 51618 51558 51198 50988 50853 50694 50589 50580 50223 50034 49458 49299
 49248 49209 48393 48204 48195 48033 48015 47529 47088 46515 45939 45915 45789
 45780 45729 45723 44994 44574 44568 44019 43995 43563 43473 43155 43038 42975
 42903 42819 42699 42633 42594 42513 42444 42153 41874 41793 41628 41289 41064
 41019 40995 40833 39948 39690 39618 39444 38850 38733 38649 38610 38568 38484
 38283 38223 38079 37959 37935 37698 36969 36804 36594 36309 35643 35523 35355
 34890 34653 34590 34383 34323 34158 33894 33858 33768 33663 33378 33375 33249
 33189 33105 32718 32430 31743 31488 31353 31215 31083 31059 30963 30903 30714

NOTE: When a random number up to a power of 2 between 1 and 2^16 is requested,
only the L1 part of the register is used. If L2 is never used, the period of
the whole random sequence doesn't exceed MFACT1*2^15-1.
------------------------------------------------------------------------------*/
#define MFACT1 65184
#define MFACT2 64860

// The maximum seed for a magic factor:
#define MFACTMAXSEED(MFACT) (((MFACT) << 16) - 2)

// Cycle the register these many times when randomizing:
#define TEMPERATION 2



/*******************************************************************************
*                                                                              *
*                                 Global Data                                  *
*                                                                              *
*******************************************************************************/

RIG_REG g_reg;    // global register



/*******************************************************************************
*                                                                              *
*                             Function Definitions                             *
*                                                                              *
*******************************************************************************/

/*******************************************************************************

FUNCTION:   GRIG_Next

PURPOSE:    Returns a random integer from the global generator.

PARAMETERS:
 max
    Maximum return value.

RETURNS:    A random integer in the range from 0 to max.

NOTES:      You must call GRIG_Randomize in your program to initialize the
            generator before using this function.

*******************************************************************************/

unsigned long GRIG_Next(unsigned long max)
{
    return RIG_Next(&g_reg, max);
}

/*******************************************************************************

FUNCTION:   GRIG_Randomize

PURPOSE:    Initializes the global random integer generator.

NOTES:      You need to call this function just one time in your program before
            you can use GRIG_NextD.

*******************************************************************************/

void GRIG_Randomize(void)
{
    RIG_Randomize(&g_reg, NULL);
}

/*******************************************************************************

FUNCTION:   RIG_Cycle, RIG16_Cycle

PURPOSE:    Cycles a register to its next state.

PARAMETERS:
 preg
    Pointer to the register to cycle.

RETURNS:    The return value is preg.

NOTES:      Use this function if you want to skip a value from a random sequence
            withot calling RIG_Next.

*******************************************************************************/

RIG_REG *__cdecl RIG_Cycle(RIG_REG *preg)
{
#ifndef _USEASM
    preg->Loops.L1 =
        LOWORD(preg->Loops.L1) * (unsigned long)MFACT1 + HIWORD(preg->Loops.L1);
    preg->Loops.L2 =
        LOWORD(preg->Loops.L2) * (unsigned long)MFACT2 + HIWORD(preg->Loops.L2);
    return preg;
#else // #ifndef _USEASM
    __asm
    {
        mov ecx, preg
        mov eax, MFACT1
        mov edx, [ecx].Loops.L1
        mul dx
        ror edx, 16
        add edx, eax
        mov [ecx].Loops.L1, edx
        mov ax, MFACT2
        mov edx, [ecx].Loops.L2
        mul dx
        ror edx, 16
        add edx, eax
        mov [ecx].Loops.L2, edx
        mov eax, ecx
    }
#endif // #ifndef _USEASM
}

/*******************************************************************************

FUNCTION:   RIG_Next

PURPOSE:    Returns the next random integer from a sequence.

PARAMETERS:
 preg
    Pointer to the register for random sequence generation. You must randomize a
    register before using it for the first time. After each call to RIG_Next,
    the register is cycled to its next state. If you randomize a register with a
    specific seed, and then call this function repeatedly with the same
    arguments, you'll get the same sequence every time.
 max
    Maximum return value.

RETURNS:    A random integer in the range from 0 to max.

*******************************************************************************/

unsigned long __cdecl RIG_Next(RIG_REG *preg, unsigned long max)
{
#ifndef _USEASM
    RIG_REG reg;
    unsigned long L1, L2;

    RIG_Cycle(preg);

    L1 = preg->Loops.L1;
    L2 = preg->Loops.L2;
    reg.Loops.L1 = (L2 << 16) | LOWORD(L1);
    reg.Loops.L2 = (L2 ^ LOWORD(L2)) | HIWORD(L1);

    return ++max ? (unsigned long)(reg.I64 % max) : reg.Loops.L1;
#else // #ifndef _USEASM
__asm
    {
        push preg
        call RIG_Cycle
        pop ebx
        mov eax, [ebx].Loops.L1
        mov ebx, [ebx].Loops.L2
        ror eax, 16
        xchg ax, bx
        ror eax, 16
        mov ecx, max
        inc ecx
        je end
        xor edx, edx
        xchg eax, ebx
        div ecx
        mov eax, ebx
        div ecx
        mov eax, edx
end:
    }
#endif // #ifndef _USEASM
}

/*******************************************************************************

FUNCTION:   RIG_Randomize

PURPOSE:    Initializes a register for random sequence generation.

PARAMETERS:
 preg
    Pointer to a register to be initialized.
 pseed
    A pointer to any data used to initialize the register in order to generate a
    specific sequence. If this parameter is NULL, the register is initialized
    randomly; otherwise, the data pointed to by pseed should be at least the
    same size as RIG_REG.

RETURNS:    The return value is preg.

*******************************************************************************/

RIG_REG *RIG_Randomize(RIG_REG *preg, const void *pseed)
{
    int i;

    if (pseed == NULL)  GetSystemTimeAsFileTime((LPFILETIME)preg);
    else                preg->I64 = *(const unsigned __int64 *)pseed;

    preg->Loops.L1 = (preg->Loops.L1 % MFACTMAXSEED(MFACT1)) + 1;
    preg->Loops.L2 = (preg->Loops.L2 % MFACTMAXSEED(MFACT2)) + 1;

    for (i = 0; i < TEMPERATION; ++i) RIG_Cycle(preg);

    return preg;
}
